import {Pipeline} from "../bones/Pipe.ts";
import {extract} from "letterparser";
import {ReplyParser} from "./replies.ts";
import DOMPurify from 'isomorphic-dompurify'
import { JSDOM } from 'jsdom'


export const allowedTags = [
    'li', 'ol', 'ul', 'ul', 'b', 'br', 'code', 'em',
    'i', 'small', 'strong', 'sub', 'sup', 'u',
]

export const lineBreakingTags = ['p', 'div']

export const sanitizeHtml = (html: string): string =>
    DOMPurify((new JSDOM('')).window).sanitize(html, { ALLOWED_TAGS: allowedTags })

/**
 * Transforms an HTML email content string by sanitizing the HTML
 * and stripping out quotes and signatures as best as possible.
 * This is imperfect and currently results in too many line breaks
 * in some cases, but it works okay enough.
 */
export const htmlReplyPipeline = Pipeline.id<string>()
    // extract the HTML from the email source
    .tap(source => extract(source).html || '')

    // here it gets weird -- the reply parser expects to work in terms of newlines and <br> tags
    // however a lot of email uses <p>. Do a pass to insert an artificial <br> at the beginning
    // of each <p> tag
    .tap(html => {
        const window = (new JSDOM('')).window
        const dp = DOMPurify(window)

        dp.addHook('uponSanitizeElement', node => {
            if ( node.nodeType !== window.Node.ELEMENT_NODE ) {
                // Skip text/document nodes, e.g.
                return
            }

            if ( node.textContent === '' && node.nodeName.toLowerCase() !== 'br' ) {
                // Drop empty nodes as long as they're not self-closing
                node.parentNode?.removeChild(node)
                return
            }

            if ( lineBreakingTags.includes(node.nodeName.toLowerCase()) ) {
                // If a wrapping tag would cause a line break, explicitly add in that break
                const br = window.document.createElement('br')
                const child = node.firstChild
                child ? node.insertBefore(br, child) : node.appendChild(br)
                return
            }
        })

        const result = dp.sanitize(html, {
            ALLOWED_TAGS: [...allowedTags, ...lineBreakingTags],
        })

        dp.removeHook('uponSanitizeElement')

        return result
    })

    // then, do a second pass which filters out the <p> tags, leaving the <br> in their place
    .tap(html => sanitizeHtml(html))

    // run that through the reply parser to strip out quotes and signatures
    .tap(html => ReplyParser.parseReply(html))

    // finally, clean it up and add the <br>'s back in since the reply parser replaced them with \n
    .tap(reply => reply.trim().replace(/\n+/g, '<br>'))
