import {getMailboxesToSearch, MailboxIterable} from "../mail/read.ts";
import {withClient} from "../mail/client.ts";
import type {Message, RootData, ThreadComment, ThreadData} from "../types.ts";
import {AsyncCollection} from "../bones/collection/AsyncCollection.ts";
import {sha256} from "../bones/crypto.ts";
import {config} from "../config.ts";
import { marked } from "marked";
import {sanitizeHtml} from "../mail/sanitize.ts";
import {formatThreadAddress} from "./id.ts";
import ColorHash from "color-hash";

export async function refreshThreadsEntirely(): Promise<void> {
    await withClient(async client => {
        const messagesByThread: {[thread: string]: Message[]} = {}

        const boxes = await getMailboxesToSearch(undefined, client)
        const now = new Date()

        for ( const box of boxes.all() ) {
            const iter = new MailboxIterable(box, client)
            const messages = new AsyncCollection(iter)

            await messages.each(message => {
                if ( !message.thread || !message.from.address ) {
                    return;
                }

                if ( !messagesByThread[message.thread] ) {
                    messagesByThread[message.thread] = []
                }

                messagesByThread[message.thread].push(message)
            })
        }

        for ( const threadId of Object.keys(messagesByThread) ) {
            const threadData: ThreadData = {
                thread: threadId,
                refresh: {
                    date: now,
                    markers: {},
                },
                comments: [],
            }

            // 3 passes required to make this work (works because objects are by-reference)
            // Pass 1: create all the ThreadComment instances and make a map by their ID
            const commentsByHash: {[hash: string]: ThreadComment} = {}
            const messages = messagesByThread[threadId]
            const colorHash = new ColorHash();
            for ( const message of messages ) {
                if (
                    !threadData.refresh.markers[message.mailbox]
                    || threadData.refresh.markers[message.mailbox] < message.modseq
                ) {
                    threadData.refresh.markers[message.mailbox] = message.modseq
                }

                commentsByHash[message.idHash] = {
                    idHash: message.idHash,
                    replyToHash: message.replyToHash,
                    replyAddress: formatThreadAddress(threadId, message.idHash),
                    user: {
                        name: message.from.name || '(anonymous)',
                        mailId: sha256(message.from.address!.toLowerCase()),
                        domainId: sha256(message.from.address!.toLowerCase().split('@').reverse()[0]),
                        color: colorHash.hex(message.from.address!.toLowerCase()),
                    },
                    date: message.date,
                    subject: message.subject,
                    text: message.content,
                    rendered: message.html || sanitizeHtml(await marked(message.content)),
                }
            }

            // Pass 2: Push replies into the replies array of their immediate parent
            for ( const message of messages ) {
                if ( !message.replyToHash ) {
                    continue
                }

                if ( !commentsByHash[message.replyToHash].replies ) {
                    commentsByHash[message.replyToHash].replies = []
                }

                commentsByHash[message.replyToHash].replies!.push(commentsByHash[message.idHash])
            }

            // Pass 3: push all top-level comments into the ThreadData
            for ( const idHash of Object.keys(commentsByHash) ) {
                const comment = commentsByHash[idHash]
                if ( comment.replyToHash ) {
                    continue
                }

                threadData.comments.push(comment)
            }

            const json = JSON.stringify(
                threadData,
                (_, v) => typeof v === 'bigint' ? `${v}` : v)
            await Bun.write(`${config.dirs.data}/threads/${threadId}.json`, json)
        }

        const root: RootData = {
            refresh: {
                date: now,
            },
        }

        await Bun.write(
            `${config.dirs.data}/root.json`,
            JSON.stringify(root),
        )
    })
}
